/**A hotspot associates Objects with each other. , both visually and programatically. 
 * They come in two flavors: Input and Output. They're connected to inputs and 
 * outputs of Objects, respectively. 
 * respectively. 
 * 
 * Usage: 
 * An Object has any number of inputs and outputs. Those are represented using 
 * Input and Output Hotspots.
 * Inputs and Outputs link to each other. (Input-Input & Output-Output links don't make sense)
 * 
 * An Input links to one or fewer Outputs, but an Output can be forked to link to any number of inputs.
 * 
 * Linked Hotspots store a common circuit state. Inputs can check the current state; 
 * outputs can check and set the state. 
 * 
 * Gates connected to the Hotspots will change and check their states as the circuit runs. 
 * 
 */
public abstract class Hotspot extends Object {
  float dx, dy;
  boolean invert = false;
  Object parent;
  color DISCONNECTED = color(0, 0, 0);

  /**Hotspot constructor
   	 * @param dx The X offset between the Object and the Hotspot
   	 * @param dy The y offset between the Object and the Hotspot
   	 */
  public Hotspot(Object obj, float dx, float dy) {
    super(dx, dy);
    name = "Hotspot";
    this.parent = obj;
  }

  public void setPen() {
    super.setPen();

    if (invert) {
      stroke(GATE);
      fill(255);
    } 
    else {
      noStroke();
      noFill();
    }
    if (!attached()) fill(DISCONNECTED);
  }

  public void draw() {
    setPen();
    float rad = 2*HOTSPOT_RADIUS_PER_GATE_SIZE * parent.getSize();
    if (parent instanceof Wire) rad = 4;
    ellipse(getX(), getY(), rad, rad);
  }

  //
  //Distance functions.
  //

  public float distance(int ox, int oy) {
    return dist(ox, oy, parent.getX()+x*parent.getSize(), parent.getY()+y*parent.getSize());
  }

  /**Whether this hotspot is in the grab radius of the specified x, y position
   	 * @param x An x-coordinate
   	 * @param y A y-coordinate
   	 * @return Whether the distance from (x, y) to this is less than GRAB_RADIUS
   	 */
  public boolean grab(int ox, int oy) {
    return HOTSPOT_GRAB_RADIUS < distance(ox, oy);
  }

  //
  //Association functions
  //

  /**Called whenever a Hotspot is moved. 
   	 * Check that all attached Hotspots are still in range, and attach any 
   	 * that've come into range.
   	 */
  public abstract void associate();

  public abstract void		attach(Hotspot hot);
  public abstract void		detach(Hotspot hot);
  public abstract boolean attached(Hotspot hot);
  public abstract boolean attached();

  /**Removes the Hotspot from the canvas.*/
  public abstract void remove();		

  /**Gets the current state of the Hotspot. 
   	 * @return The Hotspot's state.
   	 */
  public abstract boolean get();

  public Object getParent() {
    return parent;
  }

  public void setX(float x) {
    this.x = x;
    associate();
  }

  public void setY(float y) {
    this.y = y;
    associate();
  }

  public float getRealX() {
    return x;
  }

  public float getRealY() {
    return y;
  }

  public float getX() {
    return parent.getX() + parent.getSize()*x;
  }

  public float getY() {
    return parent.getY() + parent.getSize()*y;
  }

  //Override default behavior of erasing on middle click. Hotspots should never be erased on click. 
  protected void pressed() {
  }

  //Override default behavior of dragging. Hotspots are attached to gates, and should only be moved by their parent gate.
  protected void dragged() {
  }

  protected void released() {
    if (dist(mouseX, mouseY, getX(), getY()) < 
      parent.getSize() * 2 * HOTSPOT_RADIUS_PER_GATE_SIZE) negate();
  }

  public void negate() {
    //Wires can't be inverted.
    if (getParent() instanceof Wire) return;
    invert = !invert;
    dirty = true;
  } 

  public String toString() {
    return name + " at " + coord() + " attached to " + parent;
  }

  public String shortString() {
    return name + " at " + coord() + "("+parent+")";
  }
}

