/**An Object is the fundamental class that runs LogicSim (v2)*/
public abstract class Object {
  //active:	A pointer to the active object.
  //dirty:	Forces a redraw of the canvas.

  boolean selected = false;

  String name = "Object";

  Output[] output = null;
  Input[] input = null;
  float x = 0, y = 0, size = 30;

  /**The constructor must declare how many inputs there are, and 
   	 * manages whatever else comes up.
   	 * Classes that override the constructor should pass a call to super() 
   	 * to ensure proper setup.
   	 */
  public Object() {
    dirty  = true;
    input  = new Input[0];
    output = new Output[0];
    things.add(this);
  }

  /**Construct an object with an initial position.
   	 * @param x The x-coordinate
   	 * @param y The y-coordinate
   	 */
  public Object(float x, float y) {
    this();
    this.x = x;
    this.y = y;
  }

  public String toString() {
    return name+" at "+coord();
  }

  protected String coord() {
    return "(" + getX()+", "+getY()+")";
  }
  /* ==================================================================
   	 * Placement functions
   	 * ==================================================================*/

  /**@param x The new x coordinate of the gate*/
  public void setX(float x) {
    if (gridMode == OFF) this.x = x;
    else                 this.x = x - (x+GRID_SPACING/2)%GRID_SPACING + GRID_SPACING/2;
    dirty = true;
  }

  /**@param y The new y coordinate of the gate*/
  public void setY(float y) {
    if (gridMode == OFF) this.y = y;
    else                 this.y = y - (y+GRID_SPACING/2)%GRID_SPACING + GRID_SPACING/2;
    dirty = true;
  }
  /**Tries to connect the inputs and outputs of the Object that was just moved to other inputs and outputs. */
  private void reassociate() {
    for (int i = 0; i < input.length; i++) if (input[i] != null) input[i].associate();
    for (int i = 0; i < output.length; i++) if (output[i] != null) output[i].associate();
  }

  public float getX() {
    return x;
  }
  public float getY() {
    return y;
  }

  /**@param size The new size of the gate (default is 30px)*/
  public void setSize(float size) {
    this.size = size;
  }
  public float getSize() {
    return size;
  }

  /**Translates gate coordinates (-1 < x < 1; -1 < y < 1)
   	 * into absolute coordinates.
   	 * Adds the result to the current vertex list.
   	 */
  protected void tvertex(float dx, float dy) {
    vertex(x + size * dx / 2, y + size * dy / 2);
  }

  /**Translates gate coordinates (-1 < x < 1; -1 < y < 1)
   * into absolute coordinates.
   * Draws a line at the calculated position
   */
  protected void tline(float xi, float yi, float xf, float yf) {
    line(x + size * xi / 2, y + size * yi / 2, x + size * xf / 2, y + size * yf / 2);
  }

  /**Translates gate coordinates (-1 < x < 1; -1 < y < 1)
   	 * into absolute coordinates.
   	 * Draws an arc with the result.
   	 */
  protected void tarc(float a, float b, float c, float d, float start, float stop) {
    arc(x+size*a/2, y+size*b/2, c*size, d*size, start, stop);
  }

  /* ==================================================================
   	 * Simulation functions
   	 * ==================================================================*/

  /**Tick advances an Object through time. A single tick 
   	 * represents one nanosecond of real time.
   	 * 
   	 * Consider the following scenarios: an AND gate has a delay time of 10ns.
   	 * tick could wait for the inputs to remain stable for ten ns, 
   	 * then assert an output.
   	 * Or a clock could output a 10ns pulse every 100ns.
   	 * 
   	 * (Behind the scenes, the master program maintains a LinkedList of Objects, 
   	 *	and ticks each one every cycle.)
   	 */
  public abstract void tick();

  /* ==================================================================
   	 * Drawing and UI functions
   	 * ==================================================================*/

  /**draw draws the Object. Whether it's a wire stretching from a to b or an and gate. 
   	 * draw is only called when the canvas changes.
   	 * Feel free to call it conditionally from elsewhere, if ya like.
   	 */
  public abstract void draw();

  public void drawHotspots() {
    for (int i = 0; i < input.length;	i++) if (input[i]  != null) input[i].draw();
    for (int i = 0; i < output.length; i++)  if (output[i] != null) output[i].draw();
  }

  /**Sets the pen, based on the element being drawn.*/
  protected void setPen() {
    strokeWeight(3);
    stroke(0);
    noFill();
  }

  /**@return The distance between the mouse and the center of this Object.*/
  protected float distance() {
    return dist(getX(), getY(), mouseX, mouseY);
  }

  /**mousePressed forwards mousePressed events to Objects near the mouse. 
   	 * It calls the pressed() method for any object within 
   	 * MOUSE_GRAB_RADIUS of the cursor. 
   	 * In addition, if the left button is pressed, the 
   	 * first element to be selected will be marked as active and selected. 
   	 * (Only one object can be selected at a time)
   	 */
  public void mousePressed() {
    if ((this == active) || distance() < MOUSE_GRAB_RADIUS) {
      if ((active == null) || (active.distance() > distance())) {
        //If this Object is closer to the mouse than active, set this to active.
        active = this;
      }
      pressed();
    }
  }

  /**This hook is called whenever the mouse is pressed within GRAB_RADIUS of 
   	 * the Object. 
   	 * Override this within individual Object subclasses based on desired behavior. 
   	 * Important variables include mouseX, mouseY, selected, and mouseButton.
   	 */
  protected void pressed() {
    if (mouseButton == CENTER) kill.add(this);
  }

  /**MouseReleased forwards mouse release events to all nearby objects,
   	 * using the released() hook function, then deselects the previously selected object.
   	 * Note that active goes null immediately after processing the active Object. 
   	 */
  public void mouseReleased() {		
    if ((this == active) || distance() < MOUSE_GRAB_RADIUS) {
      released();
    }
    if (this == active) active = null;
  }

  /**This hook is called whenever the mouse is released within GRAB_RADIUS of 
   	 * the Object. 
   	 * Override this within individual Object subclasses based on desired behavior. 
   	 * Important variables include mouseX, mouseY, selected, and mouseButton.
   	 */
  protected void released() {
  }


  /**MouseDragged forwards mouse release events to all nearby objects,
   	 * using the dragged() hook function.
   	 */
  public void mouseDragged() {		
    if ((this == active) || distance() < MOUSE_GRAB_RADIUS) {
      dragged();
    }
  }

  /**This hook is called whenever the mouse is dragged within GRAB_RADIUS of 
   	 * the Object. 
   	 * Override this within individual Object subclasses based on desired behavior. 
   	 * Important variables include mouseX, mouseY, selected, and mouseButton.
   	 */

  protected void dragged() {
    if ((mouseButton == LEFT) && (this == active)) {
      setX(mouseX);
      setY(mouseY);
    }
    reassociate();
  }	

  /**MouseMoved just forwards mouse movement events to all nearby objects,
   	 * using the moved() hook function.
   	 */
  public void mouseMoved() {		
    if ((this == active) || distance() < MOUSE_GRAB_RADIUS) {
      moved();
    }
  }
  /**This hook is called whenever the mouse is moved within GRAB_RADIUS of 
   	 * the Object. 
   	 * Override this within individual Object subclasses based on desired behavior. 
   	 * Important variables include mouseX, mouseY, selected, and mouseButton.
   	 */
  protected void moved() {
  }

  /**Remove deletes this gate and from the canvas. It deletes the reference in 
   	 * things and detaches all inputs and outputs.
   	 */
  public void remove() {
    things.remove(this);
    for (int i = 0; i <  input.length; i++) input[i].remove();
    for (int i = 0; i < output.length; i++) output[i].remove();
    dirty = true;
  }
}

